
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "datapemcu.h"
#include "ui_datapemcu.h"
#include "log.h"

#include <QtGui>
#include <math.h>


DataPeMcu::DataPeMcu(QWidget *pParent)
:   QWidget(pParent)
    {
    m_pUi = new Ui::DataPeMcu;
    m_pUi->setupUi(this);

    m_pPeModel = new QStandardItemModel(this);
    m_pUi->viewPe->setModel(m_pPeModel);

    connect(m_pPeModel, SIGNAL(itemChanged(QStandardItem *)), this, SLOT(changePeItem(QStandardItem *)));
    }


DataPeMcu::~DataPeMcu()
    {
    delete m_pUi;
    }


void DataPeMcu::clearData()
    {
    m_pPeModel->clear();
    m_lstPeRaw.clear();
    }


bool DataPeMcu::probePeFile(const QString &qsFileName) const
    {
    LOG(4, "DataPeMcu::probePeFile(%s)", LQS(qsFileName));

    bool bIsMcuFormat = false;
    QFile oFile(qsFileName);
    oFile.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream oStream(&oFile);
    if (!oStream.atEnd())
        {
        QString qsLine = oStream.readLine();
        LOG(5, "    %s", LQS(qsLine));
        if (qsLine == QString::fromUtf8("Time, PEC Table, RA Error, DEC Error"))
            bIsMcuFormat = true;
        }
    oFile.close();

    return bIsMcuFormat;
    }


bool DataPeMcu::loadPeFile(const QString &qsFileName)
    {
    LOG(4, "DataPeMcu::loadPeFile(%s)", LQS(qsFileName));

    clearData();

    QFile oFile(qsFileName);
    oFile.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream oStream(&oFile);
    QString qsHeader = oStream.readLine();
    if (qsHeader != QString::fromUtf8("Time, PEC Table, RA Error, DEC Error"))
        {
        LOG(0, "DataPeMcu::loadPeFile(%s): wrong file format", LQS(qsFileName));
        return false;
        }
    while (!oStream.atEnd())
        {
        QString qsLine = oStream.readLine();
        LOG(5, "    %s", LQS(qsLine));
        QStringList qslEntryStrings = qsLine.split(",");
        PeRawSample oSample;
        oSample.m_oTime = QTime();
        if (qslEntryStrings.size() > 4)
            {
            // cope with files written under German language settings,
            // where decimal separators '.' becomes ','
            // heuristically convert every second ',' back to '.',
            // starting from backwards, as timestamp might have no decimals
            int nNum = 0;
            for (int nPos = qsLine.lastIndexOf(','); nPos > 0; nPos = qsLine.lastIndexOf(',', nPos-1))
                {
                if ((nNum % 2) == 0)
                    qsLine.replace(nPos, 1, '.');
                nNum++;
                }
            qslEntryStrings = qsLine.split(",");
            }
        if (qslEntryStrings.size() > 0)
            {
            oSample.m_oTime = QTime::fromString(qslEntryStrings.at(0), QString::fromUtf8("hh:mm:ss.zzz"));
            if (!oSample.m_oTime.isValid())
                oSample.m_oTime = QTime::fromString(qslEntryStrings.at(0), QString::fromUtf8("hh:mm:ss"));
            }
        oSample.m_dfWorm = (qslEntryStrings.size() > 1)? qslEntryStrings.at(1).toDouble(): 0.;
        oSample.m_dfRa = (qslEntryStrings.size() > 2)? qslEntryStrings.at(2).toDouble(): 0.;
        oSample.m_dfDec = (qslEntryStrings.size() > 3)? qslEntryStrings.at(3).toDouble(): 0.;
        m_lstPeRaw.append(oSample);
        }
    oFile.close();

    m_pPeModel->setColumnCount(4);
    m_pPeModel->setRowCount(m_lstPeRaw.size());

    QStringList qslHeaders;
    qslHeaders << tr("Time");
    qslHeaders << tr("Worm Position");
    qslHeaders << tr("RA Error");
    qslHeaders << tr("DEC Error");
    m_pPeModel->setHorizontalHeaderLabels(qslHeaders);

    for (int nRow = 0; nRow < m_lstPeRaw.size(); nRow++)
        {
        QStandardItem *pItem;
        pItem = new QStandardItem(m_lstPeRaw.at(nRow).m_oTime.toString(QString::fromUtf8("hh:mm:ss.zzz")));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPeModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(m_lstPeRaw.at(nRow).m_dfWorm, 'f', 3));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPeModel->setItem(nRow, 1, pItem);
        pItem = new QStandardItem(QString::number(m_lstPeRaw.at(nRow).m_dfRa));
        pItem->setEditable(true);
        pItem->setSelectable(true);
        m_pPeModel->setItem(nRow, 2, pItem);
        pItem = new QStandardItem(QString::number(m_lstPeRaw.at(nRow).m_dfDec));
        pItem->setEditable(true);
        pItem->setSelectable(true);
        m_pPeModel->setItem(nRow, 3, pItem);
        }

    emit dataPeChanged(this);

    return true;
    }


QList<PeRawSample> DataPeMcu::samples() const
    {
    return m_lstPeRaw;
    }


void DataPeMcu::changePeItem(QStandardItem *pItem)
    {
    LOG(5, "DataPeMcu::changePeItem");

    if (pItem == 0)
        return;
    int nCol = pItem->column();
    if ((nCol != 2) && (nCol != 3))
        return;
    int nRow = pItem->row();
    if ((nRow < 0) || (nRow >= m_lstPeRaw.size()))
        return;
    double dfValue = pItem->text().toDouble();
    if (nCol == 2)
        {
        if (m_lstPeRaw.at(nRow).m_dfRa == dfValue)
            return;
        m_lstPeRaw[nRow].m_dfRa = dfValue;
        }
    else
        {
        if (m_lstPeRaw.at(nRow).m_dfDec == dfValue)
            return;
        m_lstPeRaw[nRow].m_dfDec = dfValue;
        }

    emit dataPeChanged(this);
    }


