
/***************************************************************************
 *   This file is part of Aspect, a simple PEC tool.                       *
 *                                                                         *
 *   Copyright (C) 2006-2007 by Wolfgang Hoffmann <woho@woho.de>           *
 *                                                                         *
 *   This program is free software, licensed under the GPL v2.             *
 *   See the file COPYING for more details.                                *
 ***************************************************************************/


#include "mainwindow.h"
#include "ui_mainwindow.h"
#include "log.h"
#include "mcucmd.h"
#include "mcuserial.h"

#include <QtGui>
#include <math.h>

/*
#define DEBUG_WOHO
*/


/********************************************************************/
/*!
    \class MainWindow
    \brief Unsorted heap of all kind of stuff

    MainWindow is the GUI main window.

    \todo sort out stuff that is not related to main window
        (pe and pec handling and calculations, graphs, ...)
    \todo implement residual PE graph
    \todo replace hardcoded m_dfPecPosScale and PEC table traversal direction
*/

MainWindow::MainWindow()
:   QMainWindow()
    {
    m_pUi = new Ui::MainWindow;
    m_pUi->setupUi(this);

    m_pPeRaw = 0;
    m_aPeSorted.resize(PecData::c_nRowNum);
    m_aPeCombined.resize(PecData::c_nRowNum);
    m_aPeProcessed.resize(PecData::c_nRowNum);

    m_pPeProcessedModel = new QStandardItemModel(this);
    m_pUi->viewPeProcessed->setModel(m_pPeProcessedModel);
    m_pPecModel = new QStandardItemModel(this);
    m_pUi->viewPec->setModel(m_pPecModel);
    connect(m_pPecModel, SIGNAL(itemChanged(QStandardItem *)), this, SLOT(changePecItem(QStandardItem *)));
    m_pPecProfileModel = new QStandardItemModel(this);
    m_pUi->viewPecProfile->setModel(m_pPecProfileModel);
    m_pMcuTimer = new QTimer(this);
    connect(m_pMcuTimer, SIGNAL(timeout()), this, SLOT(mcuTimer()));

    m_pMcuPort = new McuSerial;

    m_pPeChart = new QGraphicsScene(this);
    m_pUi->chartPe->setScene(m_pPeChart);

    // todo: implement functionality
    m_pUi->toggleChartPeResidual->setChecked(false);
    m_pUi->toggleChartPeResidual->setEnabled(false);

    m_dfSiderealDay = 86164.091;
    calcPecParams(false);

/*
#ifdef DEBUG_WOHO
#else
    m_pUi->tabPecProfile->setEnabled(false);
    m_pUi->toggleChartPecProfile->setChecked(false);
    m_pUi->toggleChartPecProfile->setEnabled(false);
#endif
*/

#ifdef DEBUG_WOHO
    // woho's personal defaults to ease testing :-)
    // pixel scale: measured: 1.37, calculated: 1.31 (5.6µm @ 880mm: atan2(5.6e-6, 880e-3) / M_PI * 180. * 3600.)
    m_pUi->widgetPeIris->setPixelScale(1.31, 15. + 14./60. + 38./3600.);
    m_pUi->widgetPeIris->setWormSync(55., 91);
#if 0
    m_pUi->peProcDrift->setValue(-1.14);
    openPeFile(QString::fromUtf8("/media/sdb8/pub/webcam/pe4x.dat"));
#elif 1
    m_pUi->peProcDrift->setValue(0.);
    openPeFile(QString::fromUtf8("/home/woho/value/devel/astro/aspect/doc/pe_woho/woho_pec03.csv"));
#else
    m_pUi->peProcDrift->setValue(.3);
    openPeFile(QString::fromUtf8("/home/woho/value/devel/astro/aspect/doc/doug_dieter/pec_bad_drift.csv"));
#endif
    openPecFile(QString::fromUtf8("/media/sdb8/pub/webcam/littlefoot/pec_20061015a.csv"));
#endif

    resize(960, 640);
    }


MainWindow::~MainWindow()
    {
    delete m_pUi;
    delete m_pMcuPort;
    }


void MainWindow::on_actionQuit_triggered()
    {
    LOG(4, "MainWindow::on_actionQuit_triggered");
    close();
    }


void MainWindow::on_actionAbout_Aspect_triggered()
    {
    LOG(4, "MainWindow::on_actionAbout_Aspect_triggered");
    QMessageBox::about(this,
        tr("About Aspect ..."),
        tr("<h3>Aspect</h3>"
           "<p>Aspect is a simple PEC tool, allowing to "
           "view, edit and calculate periodic error correction "
           "tables for the <a href=\"http://littlefoot.rajiva.de/\">LittleFoot</a> "
           "astronomical mount stepper controler.</p>"
           "<p>Version: <b>0.3</b> (14 Jan 2007)<br>"
           "Autor: Wolfgang Hoffmann (<a href=\"mailto:woho@woho.de\">woho@woho.de</a>)<br>"
           "Homepage: <a href=\"http://www.woho.de/devel/astro/aspect.html\">"
           "http://www.woho.de/devel/astro/aspect.html</a></p>"
           "<p>Copyright (C) 2006-2007 Wolfgang Hoffmann.<br>"
           "This program is free software, licensed under the GPL v2.</p>"
           "<p>This program is distributed in the hope that it will be useful, "
           "but WITHOUT ANY WARRANTY; without even the implied warranty of "
           "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the "
           "GNU General Public License for more details.</p>"));
    }


void MainWindow::on_actionAbout_Qt_triggered()
    {
    LOG(4, "MainWindow::on_actionAbout_Qt_triggered");
    QApplication::aboutQt();
    }


void MainWindow::on_actionOpen_PE_Data_triggered()
    {
    LOG(4, "MainWindow::on_actionOpen_PE_Data_triggered");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptOpen);
    oFileDialog.setFileMode(QFileDialog::ExistingFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Open PE Data File"));
#ifdef DEBUG_WOHO
    oFileDialog.setDirectory("/home/woho/pub/webcam/");
#else
    oFileDialog.setDirectory(QDir::home());
#endif
    QStringList qslFilters;
    qslFilters << tr("MCU Control Plus Format (*.csv)");
    qslFilters << tr("IRIS Format (*.dat)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);

    QStringList qslFileNames;
    if (oFileDialog.exec())
        qslFileNames = oFileDialog.selectedFiles();
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        openPeFile(qsFileName);
        }
    }




/********************************************************************/
/*
    MCU connection
*/


void MainWindow::on_mcuPort_activated(const QString &qsPort)
    {
    Q_UNUSED(qsPort)
    }


void MainWindow::on_mcuAutoconnect_toggled(bool bOn)
    {
    if (!bOn)
        m_pUi->mcuAutoconnect->setChecked(true);
    }


/*!
    \todo implement
*/

void MainWindow::on_mcuConnect_toggled(bool bOn)
    {
    LOG(0, "on_mcuConnect_toggled(%s)", bOn? "on": "off");
    }


void MainWindow::on_mcuSettingsTracking_activated(int nSpeed)
    {
    McuDataTracking oTracking;
    oTracking.setRate(
        (nSpeed == 0)? McuDataTracking::sidereal:
        (nSpeed == 1)? McuDataTracking::solar:
        (nSpeed == 2)? McuDataTracking::lunar:
        McuDataTracking::off);
    if (mcuCommunicationBegin())
        {
        oTracking.mcuSet(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    }


void MainWindow::on_mcuSettingsPec_activated(int nOn)
    {
    if (nOn > 1)
        return;
    McuDataPecState oPec;
    oPec.setPec(nOn == 1);
    if (mcuCommunicationBegin())
        {
        oPec.mcuSetPec(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    }


void MainWindow::on_mcuSettingsIntelly_activated(int nOn)
    {
    McuDataPecState oPec;
    oPec.setIntellyTrack(nOn == 1);
    if (mcuCommunicationBegin())
        {
        oPec.mcuSetIntelly(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    }


void MainWindow::on_mcuSettingsRefresh_clicked()
    {
    McuDataState oState;
    McuDataPecState oPec;
    if (mcuCommunicationBegin())
        {
        oState.mcuGet(*m_pMcuPort);
        oPec.mcuGet(*m_pMcuPort);
        mcuCommunicationEnd();
        }

    m_pUi->mcuSettingsTracking->setCurrentIndex(oState.trackingRate());
    m_pUi->mcuSettingsPec->setCurrentIndex(oPec.pec());
    m_pUi->mcuSettingsPec->setEnabled(oPec.pec() < 2);
    m_pUi->mcuSettingsIntelly->setCurrentIndex(oPec.intellyTrack()? 1: 0);
    m_pUi->mcuSettingsRates->setText(
        QString::fromUtf8("2x: %1, 8x: %2, 16x: %3, Goto: %4")
        .arg(oState.guideRate())
        .arg(oState.paddle8Rate())
        .arg(oState.paddle16Rate())
        .arg(oState.gotoRate()));
    m_pUi->mcuSettingsMount->setText(oState.rotOrient());
    m_pUi->mcuSettingsGears->setText(
        QString::fromUtf8("Gear Ratio: %1, Worm Ratio: %2, Worm Wheel Teeth: %3")
        .arg(oState.gearboxReduction())
        .arg(oState.wormRatio())
        .arg(oState.wormTeeth()));
    m_pUi->mcuSettingsBacklash->setText(oState.backlash());
    }


void MainWindow::on_mcuCommandSend_clicked()
    {
    LOG(4, "MainWindow::on_mcuCommand_editingFinished");
    QByteArray aCommand = m_pUi->mcuCommand->currentText().toLatin1().constData();
    McuRawCommand oCmd;
    oCmd.setCommand(aCommand);
    if (mcuCommunicationBegin())
        {
        oCmd.mcuSend(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    QByteArray aAnswer = oCmd.answer();
    QString qsAnswer;
    for (int nByte = 0; nByte < aAnswer.length(); nByte++)
        qsAnswer += QString::fromUtf8("%1 ")
            .arg((int)((unsigned char)aAnswer.at(nByte)), 2, 16, QLatin1Char('0'));
    qsAnswer += '"';
    for (int nByte = 0; nByte < aAnswer.length(); nByte++)
        {
        QChar oChar = QChar::fromLatin1(aAnswer.at(nByte));
        qsAnswer += (oChar.isPrint())? oChar: QLatin1Char('.');
        }
    qsAnswer += '"';
    m_pUi->mcuAnswer->setText(qsAnswer);
    }


bool MainWindow::mcuCommunicationBegin()
    {
    return m_pMcuPort->open(m_pUi->mcuPort->currentText());
    }


void MainWindow::mcuCommunicationEnd()
    {
    m_pMcuPort->close();
    }




/********************************************************************/
/*
    PE raw data
*/


void MainWindow::on_widgetPeMcu_dataPeChanged(IPeRawData *pData)
    {
    LOG(4, "MainWindow::on_widgetPeMcu_dataPeChanged");
    if (m_pPeRaw == pData)
        recalcPeData();
    }


void MainWindow::on_widgetPeIris_dataPeChanged(IPeRawData *pData)
    {
    LOG(4, "MainWindow::on_widgetPeIris_dataPeChanged");
    if (m_pPeRaw == pData)
        recalcPeData();
    }




/********************************************************************/
/*
    PE processing
*/


void MainWindow::on_peProcCombining_activated(int nMethod)
    {
    LOG(4, "MainWindow::on_peProcCombining_activated");
    Q_UNUSED(nMethod)
    recalcPeProcessed();
    }


void MainWindow::on_peProcSmoothWidth_valueChanged(int nWidth)
    {
    LOG(4, "MainWindow::on_peProcSmoothWidth_valueChanged");
    Q_UNUSED(nWidth)
    recalcPeProcessed();
    }


void MainWindow::on_peProcDrift_valueChanged(double dfDrift)
    {
    LOG(4, "MainWindow::on_peProcDrift_valueChanged");
    Q_UNUSED(dfDrift)
    recalcPeData();
    }


void MainWindow::on_peProcScale_valueChanged(double dfScale)
    {
    LOG(4, "MainWindow::on_peProcScale_valueChanged");
    Q_UNUSED(dfScale)
    recalcPeData();
    }



void MainWindow::on_toggleChartPeRaw_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPeRaw_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPeProcessed_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPeProcessed_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPec_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPec_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPecProfile_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPecProfile_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::on_toggleChartPeResidual_toggled(bool bOn)
    {
    LOG(4, "MainWindow::on_toggleChartPeResidual_toggled");
    Q_UNUSED(bOn)
    recalcPeChart();
    }


void MainWindow::openPeFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::openPeFile");
    m_pUi->lineEditPeFile->setText(qsFileName);

    if (m_pPeRaw != 0)
        m_pPeRaw->clearData();

    if (m_pUi->widgetPeMcu->probePeFile(qsFileName))
        {
        m_pUi->stackPeFile->setCurrentWidget(m_pUi->pagePeMcu);
        m_pPeRaw = m_pUi->widgetPeMcu;
        }
    else if (m_pUi->widgetPeIris->probePeFile(qsFileName))
        {
        m_pUi->stackPeFile->setCurrentWidget(m_pUi->pagePeIris);
        m_pPeRaw = m_pUi->widgetPeIris;
        }
    else
        {
        m_pUi->stackPeFile->setCurrentWidget(m_pUi->pagePeNone);
        m_pPeRaw = 0;
        }

    if (m_pPeRaw != 0)
        {
        // will emit dataPeChanged() connected to recalcPeData()
        m_pPeRaw->loadPeFile(qsFileName);
        }
    else
        recalcPeData();
    }




/********************************************************************/
/*
    PEC table
*/


void MainWindow::on_actionOpen_PEC_Table_triggered()
    {
    LOG(4, "MainWindow::on_actionOpen_PEC_Table_triggered");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptOpen);
    oFileDialog.setFileMode(QFileDialog::ExistingFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Open PEC Table File"));
#ifdef DEBUG_WOHO
    oFileDialog.setDirectory("/home/woho/pub/webcam/littlefoot/");
#else
    oFileDialog.setDirectory(QDir::home());
#endif
    QStringList qslFilters;
    qslFilters << tr("PEC File (*.csv)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);

    QStringList qslFileNames;
    if (oFileDialog.exec())
        qslFileNames = oFileDialog.selectedFiles();
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        openPecFile(qsFileName);
        }
    }


void MainWindow::openPecFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::openPecFile");

    QFile oFile(qsFileName);
    oFile.open(QIODevice::ReadOnly | QIODevice::Text);
    QTextStream oStream(&oFile);
    QString qsHeader = oStream.readLine();
    if (qsHeader != QString::fromUtf8("Factor, Table, ReqSpeed, IsSpeed, MinSpeed, MaxSpeed"))
        {
        LOG(0, "MainWindow::openPecFile(): wrong file format");
        return;
        }
    m_oPecTable.clear();
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        if (oStream.atEnd())
            continue;
        QString qsLine = oStream.readLine();
        LOG(5, "    %s", qsLine.toLocal8Bit().constData());
        QStringList qslEntryStrings = qsLine.split(",");
        int nFactor = (qslEntryStrings.size() > 0)? qslEntryStrings.at(0).toInt(): 1;
        int nTable = (qslEntryStrings.size() > 1)? qslEntryStrings.at(1).toInt(): -1;
        //int nReqSpeed = (qslEntryStrings.size() > 2)? qslEntryStrings.at(2).toInt(): -1;
        //int nIsSpeed = (qslEntryStrings.size() > 3)? qslEntryStrings.at(3).toInt(): -1;
        //int nMinSpeed = (qslEntryStrings.size() > 4)? qslEntryStrings.at(4).toInt(): -1;
        //int nMaxSpeed = (qslEntryStrings.size() > 5)? qslEntryStrings.at(5).toInt(): -1;
        if (nRow == 0)
            m_oPecTable.setFactor(nFactor);
        m_oPecTable.setValue(nRow, PecSample::int2Val(nTable));
        }
    m_oPecTable.setFinish();
    oFile.close();

    fillPecModel();
    recalcPeChart();
    }


void MainWindow::on_actionSave_PEC_Table_triggered()
    {
    LOG(4, "MainWindow::on_actionSave_PEC_Table_triggered");
    QFileDialog oFileDialog(this);
    oFileDialog.setAcceptMode(QFileDialog::AcceptSave);
    oFileDialog.setFileMode(QFileDialog::AnyFile);
    oFileDialog.setViewMode(QFileDialog::Detail);
    oFileDialog.setWindowTitle(tr("Save PEC Table As File"));
#ifdef DEBUG_WOHO
    oFileDialog.setDirectory("/home/woho/pub/webcam/littlefoot/");
#else
    oFileDialog.setDirectory(QDir::home());
#endif
    QStringList qslFilters;
    qslFilters << tr("PEC File (*.csv)");
    qslFilters << tr("All Files (*)");
    oFileDialog.setFilters(qslFilters);
    oFileDialog.setDefaultSuffix(QString::fromUtf8("csv"));

    QStringList qslFileNames;
    if (oFileDialog.exec())
        qslFileNames = oFileDialog.selectedFiles();
    if (qslFileNames.size() > 0)
        {
        QString qsFileName = qslFileNames.at(0);
        LOG(3, "%s", qsFileName.toLocal8Bit().constData());
        savePecFile(qsFileName);
        }
    }


void MainWindow::savePecFile(const QString &qsFileName)
    {
    LOG(4, "MainWindow::savePecFile");

    QFile oFile(qsFileName);
    oFile.open(QIODevice::WriteOnly | QIODevice::Text);
    QTextStream oStream(&oFile);
    oStream << QString::fromUtf8("Factor, Table, ReqSpeed, IsSpeed, MinSpeed, MaxSpeed\n");
    int nReqSpeed = 0;
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        int nFactor = m_oPecTable.factor();
        int nTable = PecSample::val2Int(m_oPecTable[nRow].m_eVal);
        if (nTable == 1)
            nReqSpeed += nFactor;
        else if (nTable == 2)
            nReqSpeed -= nFactor;
        else if (nTable == 3)
            nReqSpeed = 0;
        int nIsSpeed = nReqSpeed;
        int nMinSpeed = -47;
        int nMaxSpeed = 47;
        if (nIsSpeed < nMinSpeed)
            nIsSpeed = nMinSpeed;
        if (nIsSpeed > nMaxSpeed)
            nIsSpeed = nMaxSpeed;
        // careful:
        // write CR LF to make MCU Update Plus cope with our PEC files
        oStream << QString::fromUtf8("%1, %2, %3, %4, %5, %6\r\n")
            .arg(nFactor).arg(nTable)
            .arg(nReqSpeed).arg(nIsSpeed)
            .arg(nMinSpeed).arg(nMaxSpeed);
        }
    oStream << QString::fromUtf8("\n");
    oFile.close();
    }


void MainWindow::on_actionGet_PEC_Table_triggered()
    {
    LOG(4, "MainWindow::on_actionGet_PEC_Table_triggered");
    getPecTable();
    }


/*!
    Download PEC table from MCU via serial line

    \todo fix hang if port is not available
    \todo make port configurable
    \todo evaluate return values and do some sane error handling here ...
*/

void MainWindow::getPecTable()
    {
    LOG(4, "MainWindow::getPecTable");

    QProgressDialog oProgress(tr("Downloading PEC table from MCU ..."),
        tr("Cancel"), 0, PecData::c_nRowNum, this);
    oProgress.setWindowTitle(tr("Aspect - Get PEC table"));
    oProgress.setWindowModality(Qt::WindowModal);

    McuDataPecTable oPec;
    if (mcuCommunicationBegin())
        {
        oPec.mcuGetFactor(*m_pMcuPort);
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            oProgress.setValue(nRow);
            qApp->processEvents();

            if (oProgress.wasCanceled())
                break;
            oPec.mcuGetValue(*m_pMcuPort, nRow);
            }
        oProgress.setValue(PecData::c_nRowNum);
        mcuCommunicationEnd();
        }
    m_oPecTable = oPec.data(m_oPecTable.traversal());

    fillPecModel();
    recalcPeChart();
    }


void MainWindow::on_actionPut_PEC_Table_triggered()
    {
    LOG(4, "MainWindow::on_actionPut_PEC_Table_triggered");
    putPecTable();
    }


/*!
    Upload PEC table to MCU via serial line

    \todo show message box if opening port failed
    \todo evaluate return values and do some sane error handling here ...
    \todo query and display PEC status after uploading PEC table
    \todo verify if writing order is ok (factor, save to EEPROM, PEC values, save PEC)
*/

void MainWindow::putPecTable()
    {
    LOG(4, "MainWindow::putPecTable");

    QProgressDialog oProgress(tr("Uploading PEC table to MCU ..."),
        tr("Cancel"), 0, PecData::c_nRowNum + 2, this);
    oProgress.setWindowTitle(tr("Aspect - Put PEC table"));
    oProgress.setWindowModality(Qt::WindowModal);

    McuDataPecTable oPec;
    oPec.set(m_oPecTable);

    if (mcuCommunicationBegin())
        {
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            oProgress.setValue(nRow);
            qApp->processEvents();

            if (oProgress.wasCanceled())
                break;
            oPec.mcuSetValue(*m_pMcuPort, nRow);
            }
        oProgress.setValue(PecData::c_nRowNum);
        oPec.mcuSave(*m_pMcuPort);
        oProgress.setValue(PecData::c_nRowNum + 1);
        oPec.mcuSetFactor(*m_pMcuPort);
        oProgress.setValue(PecData::c_nRowNum + 2);
        mcuCommunicationEnd();
        }
    }


void MainWindow::fillPecModel()
    {
    m_pPecModel->clear();
    m_pPecModel->setColumnCount(6);
    m_pPecModel->setRowCount(PecData::c_nRowNum);

    QStringList qslHeaders;
    qslHeaders << tr("Worm Position");
    qslHeaders << tr("PEC");
    qslHeaders << tr("acc");
    qslHeaders << tr("spd");
    qslHeaders << tr("pos");
    qslHeaders << tr("");
    m_pPecModel->setHorizontalHeaderLabels(qslHeaders);
    if (m_pUi->viewPec->header() != 0)
        m_pUi->viewPec->header()->setResizeMode(QHeaderView::ResizeToContents);

    m_pUi->pecFactor->setValue(m_oPecTable.factor());
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        QStandardItem *pItem;
        pItem = new QStandardItem(QString::number(nRow));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable.valueInt(nRow)));
        pItem->setEditable(true);
        pItem->setSelectable(true);
        m_pPecModel->setItem(nRow, 1, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable[nRow].m_nAcc));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 2, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable[nRow].m_nSpd));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 3, pItem);
        pItem = new QStandardItem(QString::number(m_oPecTable[nRow].m_nPos));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 4, pItem);
        pItem = new QStandardItem(QString());
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPecModel->setItem(nRow, 5, pItem);
        }
    }


void MainWindow::changePecItem(QStandardItem *pItem)
    {
    LOG(5, "MainWindow::changePecItem");
    if ((pItem == 0) || (pItem->column() != 1))
        return;
    int nRow = pItem->row();
    if ((nRow < 0) || (nRow >= PecData::c_nRowNum))
        return;
    int nPec = pItem->text().toInt();
    if (m_oPecTable.value(nRow) == PecSample::int2Val(nPec))
        return;
    m_oPecTable.setValue(nRow, PecSample::int2Val(nPec));
    m_oPecTable.setFinish();
    recalcPeChart();
    }


void MainWindow::on_pecFactor_valueChanged(int nFactor)
    {
    LOG(5, "MainWindow::on_pecFactor_valueChanged");
    if (m_oPecTable.factor() == nFactor)
        return;

    m_oPecTable.setFactor(nFactor);
    m_oPecTable.setFinish();
    fillPecModel();
    recalcPeChart();
    }


void MainWindow::on_pecGraphOffset_valueChanged(double dfOffset)
    {
    LOG(4, "MainWindow::on_pecGraphOffset_valueChanged");
    Q_UNUSED(dfOffset)
    recalcPeChart();
    }


/*!
    \todo show message box if opening port failed
*/

void MainWindow::on_pecMountQuery_clicked()
    {
    LOG(4, "MainWindow::on_pecMountQuery_clicked");
    McuDataState oState;
    bool bSuccess = false;
    if (mcuCommunicationBegin())
        {
        bSuccess = oState.mcuGet(*m_pMcuPort);
        mcuCommunicationEnd();
        }
    if (!bSuccess)
        {
        LOG(0, "failed to query Worm Wheel Teeth from MCU");
        return;
        }
    int nWormWheelTeeth = oState.wormTeeth();
    int nRotation = oState.rotation();
    if (nWormWheelTeeth <= 0)
        {
        LOG(0, "ignoring values, Worm Wheel Teeth %d looks strange",
            nWormWheelTeeth);
        return;
        }
    m_pUi->pecWormWheel->setValue(nWormWheelTeeth);
    m_pUi->pecRotation->setCurrentIndex(nRotation);
    }


void MainWindow::on_pecWormWheel_valueChanged(int nTeeth)
    {
    LOG(4, "MainWindow::on_pecWormWheel_valueChanged");
    Q_UNUSED(nTeeth)
    calcPecParams();
    }


void MainWindow::on_pecRotation_activated(int nDirection)
    {
    LOG(4, "MainWindow::on_pecRotation_activated");
    Q_UNUSED(nDirection)
    calcPecParams();
    }


void MainWindow::on_pecHemisphere_activated(int nHemisphere)
    {
    LOG(4, "MainWindow::on_pecHemisphere_activated");
    Q_UNUSED(nHemisphere)
    calcPecParams();
    }


void MainWindow::on_pecAccel_valueChanged(double dfAccel)
    {
    LOG(4, "MainWindow::on_pecAccel_valueChanged");
    Q_UNUSED(dfAccel)
    calcPecParams();
    }


void MainWindow::calcPecParams(bool bRedrawPecChart)
    {
    LOG(4, "MainWindow::calcPecParams");

    int nWormWheelTeeth = m_pUi->pecWormWheel->value();
    double dfAccel = m_pUi->pecAccel->value();
    double dfArcSecsPerPecPoint = 360. * 3600. / (double)nWormWheelTeeth / (double)PecData::c_nRowNum;
    double dfPecPosScale = dfAccel * dfArcSecsPerPecPoint;

    int nRotation = (m_pUi->pecRotation->currentIndex() == 0)? 1: -1;
    int nHemisphere = (m_pUi->pecHemisphere->currentIndex() == 0)? -1: 1;
    PecData::Traversal eTraversal = (nRotation * nHemisphere == 1)? PecData::forward: PecData::backward;

    bool bRedraw = bRedrawPecChart &&
        ((m_dfPecPosScale != dfPecPosScale) ||
         (m_oPecTable.traversal() != eTraversal));
    m_dfPecPosScale = dfPecPosScale;
    m_oPecTable.setTraversal(eTraversal);

    if (bRedraw)
        {
        fillPecModel();
        recalcPeChart();
        }
    }


void MainWindow::on_pecCalculate_clicked()
    {
    LOG(4, "MainWindow::on_pecCalculate_clicked");
    calcPecTable();
    }


/*!
    Calculate PEC table from processed PE data

    \sa PecData::calc()

    \todo make sure that processed PE data is valid
*/

void MainWindow::calcPecTable()
    {
    LOG(4, "MainWindow::calcPecTable");

    // convert processed PE data into a profile for compensating the PE
    QVector<double> aProfile(PecData::c_nRowNum);
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        aProfile[nRow] = - m_aPeProcessed.at(nRow) / m_dfPecPosScale;

    m_oPecTable.calc(aProfile);

    fillPecModel();
    recalcPeChart();
    }




/********************************************************************/
/*
    ...
*/


/*!
    \todo rework drift compensation
        - it relies on the presence of precise timestamps
            (it timestamp lack msecs, worm pos has better resolution, but is
            proportional to time only if PE measurement was done with PEC and IntellyTrack off)
        - it probably should consider the DEC values and work with something
            more sophisticated than linear RA scaling
*/

void MainWindow::recalcPeData()
    {
    LOG(4, "MainWindow::recalcPeData");

    m_lstPePure.clear();

    double dfDrift = m_pUi->peProcDrift->value() / 60000.;
    double dfScale = m_pUi->peProcScale->value();
    QList<PeRawSample> oData = (m_pPeRaw != 0)? m_pPeRaw->samples(): QList<PeRawSample>();
    for (int nRow = 0; nRow < oData.size(); nRow++)
        {
        PePureSample oSample;
        oSample.m_dfWorm = oData.at(nRow).m_dfWorm;
        double dfTime = oData.at(0).m_oTime.msecsTo(oData.at(nRow).m_oTime);
        oSample.m_dfPe = (oData.at(nRow).m_dfRa - dfTime * dfDrift) * dfScale;
        m_lstPePure.append(oSample);
        }

    recalcPeriods();
    }


void MainWindow::recalcPeriods()
    {
    LOG(4, "MainWindow::recalcPeriods");

    m_lstPeriod.clear();

    Period oPeriod;
    oPeriod.m_nRowBeg = -1;
    double dfWormOld = 1000.;
    for (int nRow = 0; nRow < m_lstPePure.size(); nRow++)
        {
        oPeriod.m_nRowEnd = nRow;
        double dfWormPos = m_lstPePure.at(nRow).m_dfWorm;
        // a leap of WormPos indicates end of current period
        if (fabs(dfWormPos - dfWormOld) > 128.)
            {
            if (oPeriod.m_nRowBeg >= 0)
                m_lstPeriod.append(oPeriod);
            oPeriod.m_nRowBeg = nRow;
            oPeriod.m_dfPeMin = 1e10;
            oPeriod.m_dfPeMax = -1e10;
            }
        double dfPe = m_lstPePure.at(nRow).m_dfPe;
        if (oPeriod.m_dfPeMin > dfPe)
            oPeriod.m_dfPeMin = dfPe;
        if (oPeriod.m_dfPeMax < dfPe)
            oPeriod.m_dfPeMax = dfPe;
        dfWormOld = dfWormPos;
        }
    if (oPeriod.m_nRowBeg >= 0)
        m_lstPeriod.append(oPeriod);

    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        m_aPeSorted[nRow].clear();

    for (int nRow = 0; nRow < m_lstPePure.size(); nRow++)
        {
        double dfWormPos = m_lstPePure.at(nRow).m_dfWorm;
        double dfPe = m_lstPePure.at(nRow).m_dfPe;
        int nWormPos = (int)floor(dfWormPos);
        if ((nWormPos >= 0) && (nWormPos < PecData::c_nRowNum))
            m_aPeSorted[nWormPos].append(dfPe);
        }

    recalcPeProcessed();
    }


void MainWindow::recalcPeProcessed()
    {
    LOG(4, "MainWindow::recalcPeProcessed");

    if (m_pUi->peProcCombining->currentIndex() == 0)                // 0=Average, 1=Median
        {
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            // Average
            double dfPe = 0.;
            int nSampleNum = m_aPeSorted.at(nRow).size();
            for (int nSample = 0; nSample < nSampleNum; nSample++)
                dfPe += m_aPeSorted.at(nRow).at(nSample);
            if (nSampleNum > 0)
                dfPe /= (double)nSampleNum;
            m_aPeCombined[nRow] = dfPe;
            }
        }
    else
        {
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            // Median
            double dfPe = 0.;
            int nSampleNum = m_aPeSorted.at(nRow).size();
            if (nSampleNum > 0)
                {
                QList<double> oList(m_aPeSorted.at(nRow));
                qSort(oList);
                dfPe = oList.at(nSampleNum / 2);
                }
            m_aPeCombined[nRow] = dfPe;
            }
        }

    int nSmoothWidth = m_pUi->peProcSmoothWidth->value();
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        // Average
        double dfPe = 0.;
        for (int nSmooth = 0; nSmooth < nSmoothWidth; nSmooth++)
            dfPe += m_aPeCombined.at((nRow + (nSmooth - nSmoothWidth / 2)
                + PecData::c_nRowNum) % PecData::c_nRowNum);
        m_aPeProcessed[nRow] = dfPe / (double)nSmoothWidth;
        }

    m_pPeProcessedModel->setColumnCount(2);
    m_pPeProcessedModel->setRowCount(PecData::c_nRowNum);

    QStringList qslHeaders;
    qslHeaders << tr("Worm Position");
    qslHeaders << tr("PE");
    m_pPeProcessedModel->setHorizontalHeaderLabels(qslHeaders);

    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        QStandardItem *pItem;
        pItem = new QStandardItem(QString::number(nRow));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPeProcessedModel->setItem(nRow, 0, pItem);
        pItem = new QStandardItem(QString::number(m_aPeProcessed.at(nRow)));
        pItem->setEditable(false);
        pItem->setSelectable(false);
        m_pPeProcessedModel->setItem(nRow, 1, pItem);
        }

    recalcPeChart();
    }




/********************************************************************/
/*
    PEC profile
*/


void MainWindow::on_pecProfileOffset_valueChanged(double dfOffset)
    {
    LOG(4, "MainWindow::on_pecProfileOffset_valueChanged");
    Q_UNUSED(dfOffset)
    recalcPeChart();
    }


void MainWindow::on_pecProfileDrift_valueChanged(double dfDrift)
    {
    LOG(4, "MainWindow::on_pecProfileDrift_valueChanged");
    Q_UNUSED(dfDrift)
    recalcPeChart();
    }


/*!
    \todo show message box if opening port failed
*/

void MainWindow::on_mcuStart_clicked()
    {
    LOG(4, "MainWindow::on_mcuStart_clicked");

    m_lstPecProfile.clear();

    m_pPecProfileModel->clear();
    m_pPecProfileModel->setColumnCount(3);

    QStringList qslHeaders;
    qslHeaders << tr("Time Stamp");
    qslHeaders << tr("Worm Pos");
    qslHeaders << tr("Wheel Pos");
    qslHeaders << tr("PEC Profile");
    m_pPecProfileModel->setHorizontalHeaderLabels(qslHeaders);

    // switch to sidereal tracking rate, and switch intellytrack off
    McuDataTracking oTracking;
    oTracking.setRate(McuDataTracking::sidereal);
    McuDataPecState oPec;
    oPec.setIntellyTrack(false);
    if (mcuCommunicationBegin())
        {
        oTracking.mcuSet(*m_pMcuPort);
        oPec.mcuSetIntelly(*m_pMcuPort);
        m_pMcuTimer->start((int)(m_pUi->mcuPoll->value() * 1000.));
        }
    }


void MainWindow::on_mcuStop_clicked()
    {
    m_pMcuTimer->stop();
    mcuCommunicationEnd();
    }


void MainWindow::mcuTimer()
    {
    LOG(4, "MainWindow::mcuTimer");

    McuDataPecState oPecState;
    oPecState.mcuGet(*m_pMcuPort);
    m_pUi->mcuStatusPec->setText(oPecState.state());

    PeRawSample oSample;
    oSample.m_oTime = oPecState.time();
    oSample.m_dfWorm = oPecState.wormPos();
    if (m_lstPecProfile.size() == 0)
        {
        oSample.m_dfRa = 0.;
        oSample.m_dfDec = 0.;
        }
    else
        {
        double dfWormWheelTeeth = 144.;
        double dfTime = (double)(m_lstPecProfile.at(0).m_oTime.msecsTo(oSample.m_oTime)) / 1000.;
        double dfWorm = (oSample.m_dfWorm - m_lstPecProfile.at(0).m_dfWorm) * (-1.);
        dfWorm -= floor(dfWorm / (double)PecData::c_nRowNum) * (double)PecData::c_nRowNum;
        double dfWheel = dfTime / (m_dfSiderealDay / dfWormWheelTeeth / (double)PecData::c_nRowNum);
        dfWheel -= floor(dfWheel / (double)PecData::c_nRowNum) * (double)PecData::c_nRowNum;
        double dfWormError = dfWorm - dfWheel;
        dfWormError -= floor(dfWormError / (double)PecData::c_nRowNum + .5) * (double)PecData::c_nRowNum;
        oSample.m_dfRa = (dfWormError) * (360. * 3600. / dfWormWheelTeeth / (double)PecData::c_nRowNum);
        oSample.m_dfDec = dfWheel; // abuse Dec as storage for wheel position
        }
    m_lstPecProfile.append(oSample);

    QList<QStandardItem *> oItemList;
    oItemList.append(new QStandardItem(oSample.m_oTime.toString(QString::fromUtf8("hh:mm:ss:zzz"))));
    oItemList.append(new QStandardItem(QString::number(oSample.m_dfWorm, 'f', 3)));
    oItemList.append(new QStandardItem(QString::number(oSample.m_dfDec, 'f', 3)));
    oItemList.append(new QStandardItem(QString::number(oSample.m_dfRa, 'f', 3)));
    m_pPecProfileModel->appendRow(oItemList);
    m_pUi->viewPecProfile->scrollTo(oItemList.at(0)->index());
    LOG(3, "%s %.3f %.3f %.3f",
        LQS(oSample.m_oTime.toString(QString::fromUtf8("hh:mm:ss:zzz"))),
        oSample.m_dfWorm, oSample.m_dfDec, oSample.m_dfRa);

    recalcPeChart();
    }




/********************************************************************/
/*
    Charts
*/


/*!
    Recalculate and redraw all kind of charts

    \todo move this out of MainWindow
    \todo implement PEC profile graph properly
*/

void MainWindow::recalcPeChart()
    {
    LOG(4, "MainWindow::recalcPeChart");
    QList<QGraphicsItem *> oItemList = m_pPeChart->items();
    for (int nItem = 0; nItem < oItemList.size(); nItem++)
        {
        QGraphicsItem *pItem = oItemList.at(nItem);
        m_pPeChart->removeItem(pItem);
        delete pItem;
        }

    bool bChartPeRaw = m_pUi->toggleChartPeRaw->isChecked();
    bool bChartPeProcessed = m_pUi->toggleChartPeProcessed->isChecked();
    bool bChartPec = m_pUi->toggleChartPec->isChecked();
    bool bChartPecProfile = m_pUi->toggleChartPecProfile->isChecked();
    bool bChartPeResidual = m_pUi->toggleChartPeResidual->isChecked();
    bool bChartPe = bChartPeRaw || bChartPeProcessed || bChartPeResidual;

    int nPeMin = 1000;
    int nPeMax = -1000;
    for (int nPeriod = 0; nPeriod < m_lstPeriod.size(); nPeriod++)
        {
        if (nPeMin > (int)floor(m_lstPeriod.at(nPeriod).m_dfPeMin))
            nPeMin = (int)floor(m_lstPeriod.at(nPeriod).m_dfPeMin);
        if (nPeMax < (int)ceil(m_lstPeriod.at(nPeriod).m_dfPeMax))
            nPeMax = (int)ceil(m_lstPeriod.at(nPeriod).m_dfPeMax);
        }
    int nPeZero = 0;

    double dfPecOffset = m_pUi->pecGraphOffset->value() / m_dfPecPosScale;
    int nPecMin = (int)floor(-dfPecOffset);
    int nPecMax = (int)ceil(-dfPecOffset);
    for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
        {
        double dfPos = ((double)m_oPecTable[nRow].m_nPos - dfPecOffset) * m_dfPecPosScale;
        if (nPecMin > (int)floor(dfPos))
            nPecMin = (int)floor(dfPos);
        if (nPecMax < (int)ceil(dfPos))
            nPecMax = (int)ceil(dfPos);
        }
    int nPecZero = 0;

    int nXMin = 0;
    int nXMax = PecData::c_nRowNum;
    int nYMin = 0;
    int nYMax = 0;
    if (bChartPe)
        {
        if (nYMin > nPeMin - nPeZero)
            nYMin = nPeMin - nPeZero;
        if (nYMax < nPeMax - nPeZero)
            nYMax = nPeMax - nPeZero;
        }
    if (bChartPec)
        {
        if (nYMin > -(nPecMax - nPecZero))
            nYMin = -(nPecMax - nPecZero);
        if (nYMax < -(nPecMin - nPecZero))
            nYMax = -(nPecMin - nPecZero);
        }
    int nYRange = nYMax - nYMin;
    if (nYRange < 10)
        {
        nYMin -= (10 - nYRange) / 2;
        nYMax += (10 - nYRange + 1) / 2;
        }
    double dfXScale = 500. / (double)(nXMax - nXMin + 1);
    double dfYScale = -500. / (double)(nYMax - nYMin + 1);

    double dfYScalePos = - dfYScale * m_dfPecPosScale;
    double dfYZeroSpd = 4.;
    double dfYScaleSpd = -10.;
    double dfYZeroAcc = 7.;
    double dfYScaleAcc = -10.;
    double dfYOffsPos = dfYScale * (double)nPecZero;
    double dfYOffsSpd = dfYScale * (double)nYMin - dfYScaleSpd * dfYZeroSpd;
    double dfYOffsAcc = dfYScale * (double)nYMin - dfYScaleAcc * dfYZeroAcc;

    QPainterPath oPathGrid;
    QPainterPath oPathGridMajor;
    QPainterPath oPathGridPec;
    for (int nX = nXMin; nX <= nXMax; nX += 8)
        {
        QPainterPath &rPath = ((nX % 32) == 0)? oPathGridMajor: oPathGrid;
        rPath.moveTo(dfXScale * (double)nX, dfYScale * (double)nYMin);
        rPath.lineTo(dfXScale * (double)nX, dfYScale * (double)nYMax);
        if (nX % 32 == 0)
            {
            QGraphicsTextItem *pTextItem = m_pPeChart->addText(QString::number(nX));
            pTextItem->setPos(dfXScale * (double)nX - pTextItem->boundingRect().width() / 2,
                dfYScale * (double)nYMin + 2);
            }
        if (bChartPec)
            {
            rPath.moveTo(dfXScale * (double)nX, dfYOffsSpd - dfYScaleSpd);
            rPath.lineTo(dfXScale * (double)nX, dfYOffsSpd + dfYScaleSpd);
            rPath.moveTo(dfXScale * (double)nX, dfYOffsAcc - dfYScaleAcc);
            rPath.lineTo(dfXScale * (double)nX, dfYOffsAcc + dfYScaleAcc);
            }
        }
    for (int nY = nYMin; nY <= nYMax; nY++)
        {
        QPainterPath &rPath = ((nY % 5) == 0)? oPathGridMajor: oPathGrid;
        rPath.moveTo(dfXScale * (double)nXMin, dfYScale * (double)nY);
        rPath.lineTo(dfXScale * (double)nXMax, dfYScale * (double)nY);
        if (nY % 5 == 0)
            {
            if ((bChartPe) && (nY >= nPeMin - nPeZero) && (nY <= nPeMax - nPeZero))
                {
                QGraphicsTextItem *pTextItem = m_pPeChart->addText(QString::number(nY + nPeZero));
                pTextItem->setPos(dfXScale * (double)nXMin - pTextItem->boundingRect().width() - 2,
                    dfYScale * (double)nY - pTextItem->boundingRect().height() / 2);
                }
            if ((bChartPec) && (-nY >= nPecMin - nPecZero) && (-nY <= nPecMax - nPecZero))
                {
                QGraphicsTextItem *pTextItem = m_pPeChart->addText(QString::number(- nY + nPecZero));
                pTextItem->setPos(dfXScale * (double)nXMax + 2,
                    dfYScale * (double)nY - pTextItem->boundingRect().height() / 2);
                }
            }
        }
    if (bChartPec)
        {
        oPathGrid.moveTo(dfXScale * (double)nXMin, dfYOffsSpd - dfYScaleSpd);
        oPathGrid.lineTo(dfXScale * (double)nXMax, dfYOffsSpd - dfYScaleSpd);
        oPathGrid.moveTo(dfXScale * (double)nXMin, dfYOffsSpd);
        oPathGrid.lineTo(dfXScale * (double)nXMax, dfYOffsSpd);
        oPathGrid.moveTo(dfXScale * (double)nXMin, dfYOffsSpd + dfYScaleSpd);
        oPathGrid.lineTo(dfXScale * (double)nXMax, dfYOffsSpd + dfYScaleSpd);
        oPathGrid.moveTo(dfXScale * (double)nXMin, dfYOffsAcc - dfYScaleAcc);
        oPathGrid.lineTo(dfXScale * (double)nXMax, dfYOffsAcc - dfYScaleAcc);
        oPathGrid.moveTo(dfXScale * (double)nXMin, dfYOffsAcc);
        oPathGrid.lineTo(dfXScale * (double)nXMax, dfYOffsAcc);
        oPathGrid.moveTo(dfXScale * (double)nXMin, dfYOffsAcc + dfYScaleAcc);
        oPathGrid.lineTo(dfXScale * (double)nXMax, dfYOffsAcc + dfYScaleAcc);
        }
    oPathGrid.moveTo(0., 0.);
    oPathGrid.closeSubpath();
    oPathGridMajor.moveTo(0., 0.);
    oPathGridMajor.closeSubpath();
    QGraphicsPathItem *pItem;
    pItem = m_pPeChart->addPath(oPathGrid, QPen(Qt::lightGray));
    pItem->setZValue(0.);
    pItem = m_pPeChart->addPath(oPathGridMajor, QPen(Qt::gray));
    pItem->setZValue(1.);

    if (bChartPeRaw)
        {
        for (int nPeriod = 0; nPeriod < m_lstPeriod.size(); nPeriod++)
            {
            int nRowBeg = m_lstPeriod.at(nPeriod).m_nRowBeg;
            int nRowEnd = m_lstPeriod.at(nPeriod).m_nRowEnd;
            LOG(4, "MainWindow::recalcPeChart: period %d: [%d..%d[", nPeriod, nRowBeg, nRowEnd);
            QPainterPath oPath;
            for (int nRow = nRowBeg; nRow < nRowEnd; nRow++)
                {
                double dfX = m_lstPePure.at(nRow).m_dfWorm;
                double dfY = m_lstPePure.at(nRow).m_dfPe;
                if (nRow == nRowBeg)
                    oPath.moveTo(dfXScale * dfX, dfYScale * (dfY - (double)nPeZero));
                else
                    oPath.lineTo(dfXScale * dfX, dfYScale * (dfY - (double)nPeZero));
                }
            oPath.moveTo(0., 0.);
            oPath.closeSubpath();
            pItem = m_pPeChart->addPath(oPath);
            pItem->setZValue(2.);
            }
        }

    if (bChartPeProcessed)
        {
        QPainterPath oPath;
        for (int nRow = 0; nRow <= PecData::c_nRowNum; nRow++)
            {
            double dfX = (double)nRow;
            double dfY = m_aPeProcessed.at(nRow % PecData::c_nRowNum);
            if (nRow == 0)
                oPath.moveTo(dfXScale * dfX, dfYScale * (dfY - (double)nPeZero));
            else
                oPath.lineTo(dfXScale * dfX, dfYScale * (dfY - (double)nPeZero));
            }
        oPath.moveTo(0., 0.);
        oPath.closeSubpath();
        pItem = m_pPeChart->addPath(oPath, QPen(Qt::red));
        pItem->setZValue(3.);
        }

    if (bChartPec)
        {
        QPainterPath oPathPos;
        QPainterPath oPathSpd;
        QPainterPath oPathAcc;
        oPathPos.moveTo(0., dfYOffsPos + dfYScalePos * ((double)m_oPecTable[0].m_nPos - dfPecOffset));
        oPathSpd.moveTo(0., dfYOffsSpd + dfYScaleSpd * (double)m_oPecTable[0].m_nSpd);
        oPathAcc.moveTo(0., dfYOffsAcc);
#ifdef PEC_DRAW_OLDSTYLE
        QPainterPath oPathOld;
        int nSpeed = 0;
        oPathOld.moveTo(dfXScale * ((double)nRow + .5), dfYOffsPos - dfYScale * (double)nSpeed);
#endif
        for (int nRow = 0; nRow < PecData::c_nRowNum; nRow++)
            {
            PecSample oSample(m_oPecTable[nRow]);
            double dfPos = (double)oSample.m_nPos - dfPecOffset;
            oPathPos.lineTo(dfXScale * (double)nRow, dfYOffsPos + dfYScalePos * dfPos);
            double dfSpd0 = (double)oSample.m_nSpd;
            double dfSpd1 = m_oPecTable[(nRow + 1) % PecData::c_nRowNum].m_nSpd;
            oPathSpd.lineTo(dfXScale * (double)nRow, dfYOffsSpd + dfYScaleSpd * dfSpd0);
            // need to consider traversal direction for speed
            if (m_oPecTable.traversal() == PecData::forward)
                oPathSpd.lineTo(dfXScale * ((double)nRow + 1.), dfYOffsSpd + dfYScaleSpd * dfSpd0);
            else
                oPathSpd.lineTo(dfXScale * (double)nRow, dfYOffsSpd + dfYScaleSpd * dfSpd1);
            switch (oSample.m_eVal)
                {
                case PecSample::keep:
                    break;
                case PecSample::faster:
                    oPathAcc.moveTo(dfXScale * ((double)nRow + .5), dfYOffsAcc);
                    oPathAcc.lineTo(dfXScale * ((double)nRow + .5), dfYOffsAcc + dfYScaleAcc * 1.);
                    break;
                case PecSample::slower:
                    oPathAcc.moveTo(dfXScale * ((double)nRow + .5), dfYOffsAcc);
                    oPathAcc.lineTo(dfXScale * ((double)nRow + .5), dfYOffsAcc + dfYScaleAcc * (-1.));
                    break;
                case PecSample::reset:
                    oPathAcc.moveTo(dfXScale * ((double)nRow + .5), dfYOffsAcc + dfYScaleAcc * .5);
                    oPathAcc.lineTo(dfXScale * ((double)nRow + .5), dfYOffsAcc + dfYScaleAcc * (-.5));
                    break;
                }
#ifdef PEC_DRAW_OLDSTYLE
            if (oSample.m_eVal == PecSample::reset)
                nSpeed = 0;
            else
                nSpeed += m_oPecTable[nRow].m_nAcc * m_oPecTable.factor();
            oPathOld.lineTo(dfXScale * ((double)nRow + .5), dfYOffsPos - dfYScale * (double)nSpeed);
#endif
            }
        oPathPos.lineTo(dfXScale * (double)PecData::c_nRowNum,
            dfYOffsPos + dfYScalePos * ((double)m_oPecTable[0].m_nPos - dfPecOffset));
        oPathSpd.lineTo(dfXScale * (double)PecData::c_nRowNum,
            dfYOffsSpd + dfYScaleSpd * (double)m_oPecTable[0].m_nSpd);
        oPathPos.moveTo(0., 0.);
        oPathPos.closeSubpath();
        oPathSpd.moveTo(0., 0.);
        oPathSpd.closeSubpath();
        oPathAcc.moveTo(0., 0.);
        oPathAcc.closeSubpath();
        pItem = m_pPeChart->addPath(oPathPos, QPen(Qt::darkGreen));
        pItem->setZValue(2.);
        pItem = m_pPeChart->addPath(oPathSpd, QPen(Qt::darkGreen));
        pItem->setZValue(2.);
        pItem = m_pPeChart->addPath(oPathAcc, QPen(Qt::darkGreen));
        pItem->setZValue(2.);
#ifdef PEC_DRAW_OLDSTYLE
        oPathOld.moveTo(0., 0.);
        oPathOld.closeSubpath();
        pItem = m_pPeChart->addPath(oPathOld, QPen(Qt::darkGreen));
        pItem->setZValue(2.);
#endif
        }

    if ((bChartPecProfile) && (m_lstPecProfile.size() > 0))
        {
        QPainterPath oPath;
        int nRow = (int)floor(m_lstPecProfile.at(0).m_dfWorm);
        double dfPecProfileOffset = ((double)m_oPecTable[nRow].m_nPos - dfPecOffset) * m_dfPecPosScale;
        double dfXStart = (m_lstPecProfile.size() > 0)? m_lstPecProfile.at(0).m_dfWorm: 0.;
        double dfXPrev = dfXStart;
        double dfXPeriod = 0.;
        for (int nRow = 0; nRow < m_lstPecProfile.size(); nRow++)
            {
            double dfX = m_lstPecProfile.at(nRow).m_dfWorm;
            bool bWrap = (fabs(dfX - dfXPrev) > (double)PecData::c_nRowNum / 2.);
            if (bWrap)
                dfXPeriod += (double)PecData::c_nRowNum * ((dfX > dfXPrev)? -1.: 1.);
            double dfY = dfPecProfileOffset - m_lstPecProfile.at(nRow).m_dfRa -
                (dfX + dfXPeriod - dfXStart) * m_pUi->pecProfileDrift->value() / (double)PecData::c_nRowNum -
                m_pUi->pecProfileOffset->value();
            if (bWrap || (nRow == 0))
                oPath.moveTo(dfXScale * dfX, -dfYScale * (dfY - (double)nPecZero));
            else
                oPath.lineTo(dfXScale * dfX, -dfYScale * (dfY - (double)nPecZero));
            dfXPrev = dfX;
            }
        oPath.moveTo(0., 0.);
        oPath.closeSubpath();
        pItem = m_pPeChart->addPath(oPath, QPen(QColor(170, 0, 255)));
        pItem->setZValue(3.);
        }

    if (bChartPeResidual)
        {
        }

    m_pPeChart->setSceneRect(m_pPeChart->itemsBoundingRect());
    }

